from .digitaloceanapiconnection import DigitalOceanAPIConnection
import os
import time
import queue
import threading
import datetime
import random
import json


class Snapshots(DigitalOceanAPIConnection):
    def __init__(self):
        DigitalOceanAPIConnection.__init__(self)
        self.endpoint = "/v2/snapshots"

    def list_all_snapshots(self, page=0, per_page=0):
        arguments = locals()
        del arguments["self"]
        # params must be set from a dictionary not a json dump
        params = arguments

        return self.get_request(self.endpoint, headers=self.headers, params=params)

    def list_all_droplet_snapshots(self, page=0, per_page=0):
        arguments = locals()
        del arguments["self"]
        arguments["resource_type"] = "droplet"
        # params must be set from a dictionary not a json dump
        params = arguments

        return self.get_request(self.endpoint, headers=self.headers, params=params)

    def list_all_volume_snapshots(self):
        arguments = locals()
        del arguments["self"]
        arguments["resource_type"] = "volume"
        # params must be set from a dictionary not a json dump
        params = arguments

        return self.get_request(self.endpoint, headers=self.headers, params=params)

    def retrieve_snapshot_by_id(self, id):
        return self.get_request(f"{self.endpoint}/{id}", headers=self.headers)

    def delete_snapshot_id(self, id):
        return self.delete_request(f"{self.endpoint}/{id}", headers=self.headers)



from cloudapi import BaseRESTAPI
import os


class DigitalOceanAPIConnection(BaseRESTAPI):
    def __init__(self):
        BaseRESTAPI.__init__(
            self,
            baseurl="https://api.digitalocean.com",
            callrateperhour=5000,
            # geometric_delay_multiplier=2,
            # maximum_geometric_delay_multiplications=6,
            # maximum_failed_attempts=3,
            geometric_delay_multiplier=1,
            maximum_geometric_delay_multiplications=1,
            maximum_failed_attempts=1,
        )
        if "DIGITALOCEAN_ACCESS_TOKEN" in os.environ:
            # print("Token found...")
            self.token = os.getenv("DIGITALOCEAN_ACCESS_TOKEN", "")
        else:
            raise Exception('"DIGITALOCEAN_ACCESS_TOKEN" ENV Variable not set.')
        self.headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.token} ",
        }



from .digitaloceanapiconnection import DigitalOceanAPIConnection
import os
import time
import queue
import threading
import datetime
import random
import json


class Accounts(DigitalOceanAPIConnection):
    def __init__(self):
        DigitalOceanAPIConnection.__init__(self)
        self.endpoint = "/v2/account"

    def list_account_information(self):
        return self.get_request(self.endpoint, headers=self.headers)



from .digitaloceanapiconnection import DigitalOceanAPIConnection
import os
import time
import queue
import threading
import datetime
import random
import json


class Droplets(DigitalOceanAPIConnection):
    """[summary]

    Args:
        DigitalOceanAPIConnection ([type]): [description]

    API Returns Droplet Attributes:
                id(int):	                A unique identifier for each Droplet instance.  This is automatically generated upon Droplet creation.
                name(str):	                The human-readable name set for the Droplet instance.
                memory(int):                Memory of the Droplet in megabytes.
                vcpus(int): 	            The number of virtual CPUs.
                disk(int):  	            The size of the Droplet's disk in gigabytes.
                locked(bool):	            A boolean value indicating whether the Droplet has been locked, preventing actions by users.
                created_at(str):	        A time value given in ISO8601 combined date and time format that represents when the Droplet was created.
                status(str):    	        A status string indicating the state of the Droplet instance.  This may be "new", "active", "off", or "archive".
                backup_ids([]):             An array of backup IDs of any backups that have been taken of the Droplet instance.  Droplet backups are enabled at the time of the instance creation.
                snapshot_ids([]):	        An array of snapshot IDs of any snapshots created from the Droplet instance.
                features([]):   	        An array of features enabled on this Droplet.
                region(obj):    	        The region that the Droplet instance is deployed in.  When setting a region, the value should be the slug identifier for the region.  When you query a Droplet, the entire region object will be returned.
                image(obj):     	        The base image used to create the Droplet instance.  When setting an image, the value is set to the image id or slug.  When querying the Droplet, the entire image object will be returned.
                size(obj):      	        The current size object describing the Droplet. When setting a size, the value is set to the size slug.  When querying the Droplet, the entire size object will be returned. Note that the disk volume of a Droplet may not match the size's disk due to Droplet resize actions. The disk attribute on the Droplet should always be referenced.
                size_slug(str): 	        The unique slug identifier for the size of this Droplet.
                networks(obj):  	        The details of the network that are configured for the Droplet instance.  This is an object that contains keys for IPv4 and IPv6.  The value of each of these is an array that contains objects describing an individual IP resource allocated to the Droplet.  These will define attributes like the IP address, netmask, and gateway of the specific network depending on the type of network it is.
                kernel(obj):    	        Nullable object	The current kernel. This will initially be set to the kernel of the base image when the Droplet is created.
                next_backup_window(obj)	    Nullable object	The details of the Droplet's backups feature, if backups are configured for the Droplet. This object contains keys for the start and end times of the window during which the backup will start.
                tags([]):	                An array of Tags the Droplet has been tagged with.
                volume_ids([]):             A flat array including the unique identifier for each Block Storage volume attached to the Droplet.
                vpc_uuid([]):           	A string specifying the UUID of the VPC to which the Droplet is assigned.
    """

    def __init__(self):
        DigitalOceanAPIConnection.__init__(self)
        self.endpoint = "/v2/droplets"

    def list_all_droplets(self, page=0, per_page=0):
        """[summary]

        Returns:
            [type]: [description]
        API Expects:
                page(int)                   which page to return
                per_page(int)               how many results per page
        API Returns:
                A list of droplets objects, with standard droplet attributes.

        """
        arguments = locals()
        del arguments["self"]
        # params must be set from a dictionary not a json dump
        params = arguments

        return self.get_request(self.endpoint, headers=self.headers, params=params)

    def list_all_droplets_by_tag(self, tag_name, page=0, per_page=0):
        arguments = locals()
        del arguments["self"]
        # params must be set from a dictionary not a json dump
        params = arguments

        return self.get_request(self.endpoint, headers=self.headers, params=params)

    def create_new_droplet(
        self,
        name,
        region,
        size,
        image,
        ssh_keys=[],
        backups=None,
        ipv6=None,
        private_networking=None,
        vpc_uuid=None,
        user_data=None,
        monitoring=None,
        volumes=[],
        tags=[],
    ):

        """[summary]

        Returns:
            [type]: [description]
        API Expects:
                name(str):	                REQUIRED.  The human-readable string you wish to use when displaying the Droplet name. The name, if set to a domain name managed in the DigitalOcean DNS management system, will configure a PTR record for the Droplet. The name set during creation will also determine the hostname for the Droplet in its internal configuration.
                region(str):	            REQUIRED.  The unique slug identifier for the region that you wish to deploy in.
                size(str):                  REQUIRED.  The unique slug identifier for the size that you wish to select for this Droplet.
                image(int,str):	            REQUIRED.  integer (if using an image ID), or String (if using a public image slug)	The image ID of a public or private image, or the unique slug identifier for a public image. This image will be the base image for your Droplet.
                ssh_keys([]):	            An array containing the IDs or fingerprints of the SSH keys that you wish to embed in the Droplet's root account upon creation.
                backups(bool):              A boolean indicating whether automated backups should be enabled for the Droplet. Automated backups can only be enabled when the Droplet is created.
                ipv6(bool):                 A boolean indicating whether IPv6 is enabled on the Droplet.
                private_networking(bool):	This parameter has been deprecated. Use 'vpc_uuid' instead to specify a VPC network for the Droplet. If no `vpc_uuid` is provided, the Droplet will be placed in the default VPC.
                vpc_uuid(str):          	A string specifying the UUID of the VPC to which the Droplet will be assigned. If excluded, beginning on April 7th, 2020, the Droplet will be assigned to your account's default VPC for the region.
                user_data(str):             A string containing 'user data' which may be used to configure the Droplet on first boot, often a 'cloud-config' file or Bash script. It must be plain text and may not exceed 64 KiB in size.
                monitoring(bool):           A boolean indicating whether to install the DigitalOcean agent for monitoring.
                volumes([]):            	A flat array including the unique string identifier for each block storage volume to be attached to the Droplet. At the moment a volume can only be attached to a single Droplet.
                tags([])                    A flat array of tag names as strings to apply to the Droplet after it is created. Tag names can either be existing or new tags.

        API Returns:
               A droplets object, with standard droplet attributes.

        API data example:
                data = '{"name":"example.com","region":"nyc3","size":"s-1vcpu-1gb","image":"ubuntu-16-04-x64","ssh_keys":[107149],"backups":false,"ipv6":true,"user_data":null,"private_networking":null,"volumes": null,"tags":["web"]}'


        """
        arguments = locals()
        del arguments["self"]
        data = json.dumps(arguments)
        return self.post_request(self.endpoint, headers=self.headers, data=data)

    def delete_droplet_id(self, id):
        """
        To delete a Droplet, send a DELETE request to /v2/droplets/$DROPLET_ID
        """
        return self.delete_request(f"{self.endpoint}/{id}", headers=self.headers)

    def delete_droplet_tag(self, tag_name):
        """
        To delete Droplets by a tag (for example awesome), send a DELETE request to /v2/droplets?tag_name=$TAG_NAME.
        """
        return self.delete_request(
            f"{self.endpoint}?tag_name={tag_name}", headers=self.headers
        )

    def retrieve_droplet_by_id(self, id):
        """[summary]

        Args:
            id ([type]): [description]
            To show information about an individual Droplet, send a GET request to /v2/droplets/$DROPLET_ID.
        """
        return self.get_request(f"{self.endpoint}/{id}", headers=self.headers)

    def retrieve_droplet_action(self, droplet_id, action_id):
        """

        Args:
            droplet_id ([type]): [description]
            action_id ([type]): [description]
        """

        return self.get_request(
            f"{self.endpoint}/{droplet_id}/actions/{action_id}", headers=self.headers
        )

    def reboot_droplet(self, id):
        """

        Args:
            id ([type]): [description]
        """
        data_dict = {}
        data_dict["type"] = "reboot"
        data = json.dumps(data_dict)
        return self.post_request(
            f"{self.endpoint}/{id}/actions", headers=self.headers, data=data
        )

    def shutdown_droplet(self, id):
        """

        Args:
            id ([type]): [description]
        """
        data_dict = {}
        data_dict["type"] = "shutdown"
        data = json.dumps(data_dict)
        return self.post_request(
            f"{self.endpoint}/{id}/actions", headers=self.headers, data=data
        )

    def poweron_droplet(self, id):
        """

        Args:
            id ([type]): [description]
        """
        data_dict = {}
        data_dict["type"] = "power_on"
        data = json.dumps(data_dict)
        return self.post_request(
            f"{self.endpoint}/{id}/actions", headers=self.headers, data=data
        )

    def poweroff_droplet(self, id):
        """

        Args:
            id ([type]): [description]
        """
        data_dict = {}
        data_dict["type"] = "power_off"
        data = json.dumps(data_dict)
        return self.post_request(
            f"{self.endpoint}/{id}/actions", headers=self.headers, data=data
        )

    def powercycle_droplet(self, id):
        """

        Args:
            id ([type]): [description]
        """
        data_dict = {}
        data_dict["type"] = "power_cycle"
        data = json.dumps(data_dict)
        return self.post_request(
            f"{self.endpoint}/{id}/actions", headers=self.headers, data=data
        )

    def rebuild_droplet(self, id, image):
        """
        Rebuld action functions just like a new create.

        Args:
            id ([type]): Droplet ID
            image ([type]): Image slug or ID.

        Returns:
            [type]: [description]
        """
        data_dict = {}
        data_dict["type"] = "rebuild"
        data_dict["image"] = image
        data = json.dumps(data_dict)
        return self.post_request(
            f"{self.endpoint}/{id}/actions", headers=self.headers, data=data
        )

    def rename_droplet(self, id, name):
        print("rename api called")
        """
        Rebuld action functions just like a new create.

        Args:
            id ([type]): Droplet ID
            name ([type]): New droplet name

        Returns:
            [type]: [description]
        """
        data_dict = {}
        data_dict["type"] = "rename"
        data_dict["name"] = name
        data = json.dumps(data_dict)
        endpoint = f"{self.endpoint}/{id}/actions"
        print(endpoint)
        print(data)
        return self.post_request(endpoint, headers=self.headers, data=data)
        print("api finished")

    def create_snapshot_from_droplet(self, id, name):
        data_dict = {}
        data_dict["type"] = "snapshot"
        data_dict["name"] = name
        # api doesnt let you have tags for droplets at this point
        # data_dict["tags"] = tags
        data = json.dumps(data_dict)
        return self.post_request(
            f"{self.endpoint}/{id}/actions", headers=self.headers, data=data
        )

    def list_snapshots_for_droplet(self, id, page=0, per_page=0):
        arguments = locals()
        del arguments["self"]
        del arguments["id"]
        # params must be set from a dictionary not a json dump
        params = arguments
        return self.get_request(
            f"{self.endpoint}/{id}/snapshots", headers=self.headers, params=params
        )

    def restore_droplet(self, droplet_id, image_id):
        data_dict = {}
        data_dict["type"] = "restore"
        data_dict["image"] = image_id
        data = json.dumps(data_dict)

        return self.post_request(
            f"{self.endpoint}/{droplet_id}/actions", headers=self.headers, data=data
        )

    def resize_droplet(self, droplet_id, size, disk_resize=False):
        ##WARNING... if you try to resie to a lower size disk  api will hang.
        ##need to implement checking if resize is allowed.
        """[summary]

        Args:
            droplet_id ([type]): [description]
            size ([type]): [description]
            disk_resize (bool, optional): When set to True you resize not only the memory, but also you upgrade to the disk size of the slug, meaning you cant resize back down. Defaults to False.

        Returns:
            [type]: [description]
        """
        data_dict = {}
        data_dict["type"] = "resize"
        if disk_resize:
            data_dict["disk"] = True
        data_dict["size"] = size
        data = json.dumps(data_dict)

        return self.post_request(
            f"{self.endpoint}/{droplet_id}/actions", headers=self.headers, data=data
        )

    def list_droplet_resources(self, droplet_id):
        return self.get_request(
            f"{self.endpoint}/{droplet_id}/destroy_with_associated_resources",
            headers=self.headers,
        )


if __name__ == "__main__":
    digitalocean_droplets = Droplets()

    # def make_a_call_to_digitalocea_to_list_all_droplets(x):
    #    response = digitalocean_droplets.list_all_droplets()
    #    print(x, datetime.datetime.now())

    # for x in range(0, 20):
    #    threading.Thread(
    #        target=make_a_call_to_digitalocea_to_list_all_droplets, args=(x,)
    #    ).start()

    response = digitalocean_droplets.create_new_droplet(
        name="example.com",
        region="nyc3",
        size="s-1vcpu-1gb",
        image="ubuntu-16-04-x64",
        ssh_keys=[],
        backups=False,
        ipv6=True,
        user_data=None,
        private_networking=None,
        volumes=None,
        tags=["banabas"],
    )

    # response = digitalocean_droplets.delete_droplet_id(249612802)

    # response = digitalocean_droplets.list_all_droplets(page=1, per_page=2)
    content = response.content.decode("utf-8")
    droplet_data = dict(json.loads(content)["droplet"])
    print(droplet_data)



from __future__ import annotations

from dataclasses import dataclass, field
from ..digitaloceanapi.droplets import Droplets
from ..digitaloceanapi.volumes import Volumes
from .action import *
from .snapshot import *
from .size import *
from .volume import *
from .account import *
from ..common.cloudapiexceptions import *
import json
import threading
import time
import re


@dataclass
class DropletSnapshotAttributes:
    id: int = None
    name: str = None
    distribution: str = None
    slug: str = None
    public: bool = None
    regions: list = field(default_factory=list)
    created_at: str = None
    min_disk_size: int = None
    type: str = None
    size_gigabytes: float = None


class DropletSnapshot:
    def __init__(self):
        self.attributes = DropletSnapshotAttributes()


class DropletManager:
    def __init__(self):
        self.dropletapi = Droplets()
        self.smanager = SnapshotManager()
        self.amanager = ActionManager()
        self.account_manager=AccountManager()

    def check_limit(self):
        droplet_limit = self.account_manager.droplet_limit()
        if not len(self.retrieve_all_droplets()) < droplet_limit:
            raise ErrorAccountDropletLimitReached
            (
                f"You have reached your droplet limit of {droplet_limit}"
            )

    def is_valid_droplet_name(self,droplet_name):
        #Double check hostname for valid chars
        re_is_valid_hostname=re.compile('^[a-zA-Z0-9.-]+$').search
        if not bool(re_is_valid_hostname(droplet_name)):
            raise ErrorDropletNameContainsInvalidChars(f"\"{droplet_name}\" is not a valid hostname, droplet names must contain only (a-z, A-Z, 0-9, . and -)")


    def create_new_droplet(
        self,
        name,
        region,
        size,
        image,
        ssh_keys=[],
        backups=None,
        ipv6=None,
        private_networking=None,
        vpc_uuid=None,
        user_data=None,
        monitoring=None,
        volumes=[],
        tags=[],
    ):
        arguments = locals()
        del arguments["self"]


        self.is_valid_droplet_name(arguments['name'])
        self.check_limit()



        newdroplet = Droplet()
        newdroplet.arguments = DropletArguments(**arguments)
        response = self.dropletapi.create_new_droplet(**arguments)
        if response:
            #
            droplet_data = dict(json.loads(response.content.decode("utf-8"))["droplet"])
            newdroplet.attributes = DropletAttributes(**droplet_data)
        else:
            raise Exception(f"Could not create droplet {name}, {response.content}")
        return newdroplet

    def retrieve_droplet_by_id(self, id):
        """
        Returns a Droplet object containing attributes for a droplet with id.

        Args:
            id ([type]): [description]

        Returns:
            [Droplet]:A droplet object containing attributes for a droplet with object id.
        """
        if not self.does_droplet_id_exist(id):
            raise ErrorDropletNotFound(f"Droplet with id:{id} does not exists")
        newdroplet = Droplet(status="retrieve")
        response = self.dropletapi.retrieve_droplet_by_id(id)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            droplet_data = dict(content["droplet"])
            newdroplet.attributes = DropletAttributes(**droplet_data)
        return newdroplet

    def retrieve_droplets_by_name(self, name):
        return_droplets = []
        droplets = self.retrieve_all_droplets()
        for droplet in droplets:
            if droplet.attributes.name == name:
                return_droplets.append(droplet)
        return return_droplets

    def retrieve_all_droplets(self):
        """
        Returns an array of Droplet objects, one for each droplet in digitalocean account.

        Returns:
            [type]: [description]
        """

        # Build list of droplets from api, but take in to account possible pagination.
        droplet_list = []
        page, per_page = 1, 10
        response = self.dropletapi.list_all_droplets(page=page, per_page=per_page)
        content = json.loads(response.content.decode("utf-8"))
        droplet_list.extend(content["droplets"])
        try:
            while content["links"]["pages"]["next"]:
                page = page + 1
                response = self.dropletapi.list_all_droplets(
                    page=page, per_page=per_page
                )
                content = json.loads(response.content.decode("utf-8"))
                droplet_list.extend(content["droplets"])
        except KeyError:
            pass

        # Build and return that Droplet object array.
        droplet_objects = []
        for droplet_item in droplet_list:
            newdroplet = Droplet(status="retrieve")
            newdroplet.attributes = DropletAttributes(**droplet_item)
            droplet_objects.append(newdroplet)
        return droplet_objects

    # def retrieve_all_droplets_by_tag(self, tag_name=None):
    #    """
    #    Returns an array of Droplet objects, one for each droplet in digitalocean account.
    #
    #    Returns:
    #        [type]: [description]
    #    """
    #
    # If no tag has been provided, return no droplets
    #    if tag_name == None:
    #        return []

    # Build list of droplets from api, but take in to account possible pagination.
    #    droplet_list = []
    #    page, per_page = 1, 10
    #    response = self.dropletapi.list_all_droplets_by_tag(
    #        tag_name=tag_name, page=page, per_page=per_page
    #    )
    #    content = json.loads(response.content.decode("utf-8"))
    #    droplet_list.extend(content["droplets"])
    #    try:
    #        while content["links"]["pages"]["next"]:
    #            page = page + 1
    #            response = self.dropletapi.list_all_droplets(
    #                page=page, per_page=per_page
    #            )
    #            content = json.loads(response.content.decode("utf-8"))
    #            droplet_list.extend(content["droplets"])
    #    except KeyError:
    #        pass

    # Build and return that Droplet object array.
    #    droplet_objects = []
    #    for droplet_item in droplet_list:
    #        newdroplet = Droplet(status="retrieve")
    #        newdroplet.attributes = DropletAttributes(**droplet_item)
    #        droplet_objects.append(newdroplet)
    #    return droplet_objects

    def retrieve_droplets_with_only_tags(self, tag: list):
        if isinstance(tag, str):
            tag = [tag]
        return_droplets = []
        droplets = self.retrieve_all_droplets()
        for droplet in droplets:
            if set(list(tag)) == set(droplet.attributes.tags):
                return_droplets.append(droplet)
        if len(return_droplets) > 0:
            return return_droplets
        else:
            raise ErrorDropletNotFound(
                f"No droplets containing all tags: {tag} were found."
            )

    def delete_droplets_with_only_tags(self, tag: list):
        if isinstance(tag, str):
            tag = [tag]
        try:
            droplets = self.retrieve_droplets_with_only_tags(tag)
            for droplet in droplets:
                self.delete_droplet(droplet)
        except:
            pass

    def retrieve_droplets_with_all_tags(self, tag: list):
        if isinstance(tag, str):
            tag = [tag]
        return_droplets = []
        droplets = self.retrieve_all_droplets()
        for droplet in droplets:
            if set(list(tag)).issubset(set(droplet.attributes.tags)):
                return_droplets.append(droplet)
        if len(return_droplets) > 0:
            return return_droplets
        else:
            raise ErrorDropletNotFound(
                f"No droplets containing all tags: {tag} were found."
            )

    def delete_droplets_with_all_tags(self, tag: list):
        if isinstance(tag, str):
            tag = [tag]
        try:
            droplets = self.retrieve_droplets_with_all_tags(tag)
            for droplet in droplets:
                self.delete_droplet(droplet)
        except:
            pass

    def retrieve_droplets_with_any_tags(self, tag: list):
        if isinstance(tag, str):
            tag = [tag]
        return_droplets = []
        droplets = self.retrieve_all_droplets()
        for droplet in droplets:
            if not set(list(tag)).isdisjoint(set(droplet.attributes.tags)):
                return_droplets.append(droplet)
        if len(return_droplets) > 0:
            return return_droplets
        else:
            raise ErrorDropletNotFound(
                f"No droplets containing all tags: {tag} were found."
            )

    def delete_droplets_with_any_tags(self, tag: list):
        if isinstance(tag, str):
            tag = [tag]
        try:
            droplets = self.retrieve_droplets_with_any_tags(tag)
            for droplet in droplets:
                self.delete_droplet(droplet)
        except:
            pass

    def delete_droplet(self, droplet: Droplet):
        if not droplet.deleted==False:
            raise ErrorDropletNotFound(f"{droplet.attributes.id} was already deleted")
        self.delete_droplet_by_id(droplet.attributes.id)
        droplet.deleted=True 

    def delete_droplet_by_id(self, id):
        if self.does_droplet_id_exist(id):
            response = self.dropletapi.delete_droplet_id(id)

    #def delete_droplets_by_tag(self, tag_name=None):
    #    if not tag_name == None:
    #        response = self.dropletapi.delete_droplet_tag(tag_name=tag_name)

    
    def does_droplet_id_exist(self, id):
        droplets = self.retrieve_all_droplets()
        for droplet in droplets:
            if str(droplet.attributes.id) == str(id):
                return True
        return False


@dataclass
class DropletAttributes:
    id: int = None
    name: str = None
    memory: int = None
    vcpus: int = None
    disk: int = None
    locked: bool = None
    created_at: str = None
    status: str = None
    backup_ids: list = field(default_factory=list)
    snapshot_ids: list = field(default_factory=list)
    features: list = field(default_factory=list)
    region: object = field(default_factory=list)
    image: object = field(default_factory=list)
    size: object = field(default_factory=list)
    size_slug: str = None
    networks: object = field(default_factory=list)
    kernel: object = field(default_factory=list)
    next_backup_window: object = field(default_factory=list)
    tags: list = field(default_factory=list)
    volume_ids: list = field(default_factory=list)
    vpc_uuid: list = field(default_factory=list)


@dataclass
class DropletArguments:

    name: str = None
    region: str = None
    size: str = None
    image: object = None
    ssh_keys: list = field(default_factory=list)
    backups: bool = None
    ipv6: bool = None
    private_networking: bool = None
    vpc_uuid: str = None
    user_data: str = None
    monitoring: bool = None
    volumes: list = field(default_factory=list)
    tags: list = field(default_factory=list)


#@dataclass
#class DropletLastAction:
#    id: int = None
#    status: str = None
#    type: str = None
#    started_at: str = None
#    completed_at: str = None
#    resource_id: int = None
#    resource_type: str = None
#    region: object = None
#    region_slug: str = None


class Droplet:
    def __init__(self, status=None):
        self.arguments = DropletArguments()
        self.attributes = DropletAttributes()
        self.lastaction:Action = None
        self.attributes.status = status
        self.dropletapi = Droplets()
        self.volumeapi = Volumes()
        self.action_manager = ActionManager()
        self.size_manager = SizeManager()
        self.volume_manager = VolumeManager()
        self.snapshot_manager = SnapshotManager()
        self.droplet_manager = DropletManager()
        self.update_on_active_status()
        self.deleted=False

    def update(self):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        """
        Updates the Droplet attributes data class with the latest droplet information at digital ocean.
        """
        response = self.dropletapi.retrieve_droplet_by_id(self.attributes.id)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            droplet_data = dict(content["droplet"])
            self.attributes = DropletAttributes(**droplet_data)

    def update_on_active_status(self):
        """
        A freshly created droplet will need time to completely boot up and be active.
        Information like IP addresses are not available untill the droplet is active.
        Here we start a background thread that waits for the droplet to become active and then update the droplet attributes.
        """

        def update_status():
            while not self.attributes.status == "active":
                if self.attributes.status == None:
                    time.sleep(10)
                elif self.attributes.status == "new":
                    self.update()
                    time.sleep(10)
                else:
                    break

        thread = threading.Thread(target=update_status, args=())
        thread.start()

    ###### Do we need to update droplet action, if action already updates itself

    # def update_droplet_action(self):
    #    """
    #    Updates the Droplet lastaction data class with the latest droplet action information at digital ocean.
    #    """
    #    response = self.dropletapi.retrieve_droplet_action(
    #        self.attributes.id, self.lastaction.id
    #    )
    #    if response:
    #        content = json.loads(response.content.decode("utf-8"))
    #        action_data = dict(content["action"])
    #        self.lastaction = DropletLastAction(**action_data)

    # def update_on_active_action(self):
    #    """
    #    A freshly created droplet will need time to completely boot up and be active.
    #    Information like IP addresses are not available untill the droplet is active.
    #    Here we start a background thread that waits for the droplet to become active and then update the droplet attributes.
    #    """

    #    def update_action():
    #        while (self.lastaction.status == None) or (
    #            self.lastaction.status == "in-progress"
    #        ):
    #            if not self.lastaction.status in ["completed", "errored"]:
    #                time.sleep(10)
    #                self.update_droplet_action()
    #            else:
    #                break
    #            if self.lastaction.status == "errored":
    #                raise ErrorActionFailed(
    #                    f"Action {self.attributes.id},{self.attributes.type} failed"
    #                )
    #
    #    thread = threading.Thread(target=update_action, args=())
    #    thread.start()

    def reboot(self):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        droplet_id = self.attributes.id
        response = self.dropletapi.reboot_droplet(droplet_id)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            action_data = dict(content["action"])
            newaction = Action()
            newaction.attributes = ActionAttributes(**action_data)
            self.action_manager.wait_for_action_completion(newaction)
            self.lastaction = newaction

    def powercycle(self):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        droplet_id = self.attributes.id
        response = self.dropletapi.powercycle_droplet(droplet_id)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            action_data = dict(content["action"])
            newaction = Action()
            newaction.attributes = ActionAttributes(**action_data)
            self.action_manager.wait_for_action_completion(newaction)
            self.lastaction = newaction

    def shutdown(self):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        droplet_id = self.attributes.id
        response = self.dropletapi.shutdown_droplet(droplet_id)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            action_data = dict(content["action"])
            newaction = Action()
            newaction.attributes = ActionAttributes(**action_data)
            self.action_manager.wait_for_action_completion(newaction)
            self.lastaction = newaction

    def poweroff(self):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        droplet_id = self.attributes.id
        response = self.dropletapi.poweroff_droplet(droplet_id)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            action_data = dict(content["action"])
            newaction = Action()
            newaction.attributes = ActionAttributes(**action_data)
            self.action_manager.wait_for_action_completion(newaction)
            self.lastaction = newaction

    def poweron(self):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        droplet_id = self.attributes.id
        response = self.dropletapi.poweron_droplet(droplet_id)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            action_data = dict(content["action"])
            newaction = Action()
            newaction.attributes = ActionAttributes(**action_data)
            self.action_manager.wait_for_action_completion(newaction)
            self.lastaction = newaction

    def rebuild(self, image):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        droplet_id = self.attributes.id
        response = self.dropletapi.rebuild_droplet(droplet_id, image)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            action_data = dict(content["action"])
            newaction = Action()
            newaction.attributes = ActionAttributes(**action_data)
            self.action_manager.wait_for_action_completion(newaction)
            self.lastaction = newaction

    def rename(self, name):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        self.droplet_manager.is_valid_droplet_name(name)
        droplet_id = self.attributes.id
        response = self.dropletapi.rename_droplet(droplet_id, name)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            action_data = dict(content["action"])
            newaction = Action()
            newaction.attributes = ActionAttributes(**action_data)
            self.action_manager.wait_for_action_completion(newaction)
            self.lastaction = newaction

    def create_snapshot(self, name):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        id = self.attributes.id
        response = self.dropletapi.create_snapshot_from_droplet(id, name)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            action_data = dict(content["action"])
            newaction = Action(ActionAttributes(**action_data))
            self.action_manager.wait_for_action_completion(newaction)
            self.lastaction = newaction
            #print(newaction.attributes.started_at)
            #print(newaction.attributes.completed_at)
            snapshot_objects=self.retrieve_snapshots()
            for snapshot_object in snapshot_objects:
                if (snapshot_object.attributes.name==name) and (snapshot_object.attributes.created_at==newaction.attributes.started_at):
                    return snapshot_object
            return None 

    def restore_droplet(self, image_id):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        id = self.attributes.id
        response = self.dropletapi.restore_droplet(id, image_id)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            action_data = dict(content["action"])
            newaction = Action(ActionAttributes(**action_data))
            self.action_manager.wait_for_action_completion(newaction)
            self.lastaction = newaction

    def resize_droplet(self, slug_size, disk_resize=False):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        # OK, if you try and resize to a smaller disk you will fail.
        desired_size = self.size_manager.retrieve_size(slug_size)
        target_disk_size = desired_size.attributes.disk
        current_disk_size = self.attributes.disk
        if target_disk_size < current_disk_size:
            raise ErrorDropletResizeDiskError(
                "You can't resize to a smaller disk, resize to same disk size with different RAM memory instead"
            )

        id = self.attributes.id
        response = self.dropletapi.resize_droplet(
            id, slug_size, disk_resize=disk_resize
        )
        if response:
            content = json.loads(response.content.decode("utf-8"))
            action_data = dict(content["action"])
            newaction = Action()
            newaction.attributes = ActionAttributes(**action_data)
            self.action_manager.wait_for_action_completion(newaction)
            self.lastaction = newaction

            self.poweron()

    def delete(self):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        self.dropletapi.delete_droplet_id(self.attributes.id)
        self.deleted=True

    def retrieve_snapshots(self):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        id = self.attributes.id
        snapshot_list = []
        page, per_page = 1, 10
        response = self.dropletapi.list_snapshots_for_droplet(
            id=id, page=page, per_page=per_page
        )
        content = json.loads(response.content.decode("utf-8"))
        snapshot_list.extend(content["snapshots"])
        try:
            while content["links"]["pages"]["next"]:
                page = page + 1
                response = self.dropletapi.list_snapshots_for_droplet(
                    id=id, page=page, per_page=per_page
                )
                content = json.loads(response.content.decode("utf-8"))
                snapshot_list.extend(content["snapshots"])
        except KeyError:
            pass
            #print(snapshot_list)
        # Build and return that Snapshot object array.
        dropletsnapshot_objects = []
        for snapshot_item in snapshot_list:
            newdropletsnapshot = DropletSnapshot()
            newdropletsnapshot.attributes = DropletSnapshotAttributes(**snapshot_item)
            dropletsnapshot_objects.append(newdropletsnapshot)
        return dropletsnapshot_objects

    def retrieve_snapshot_by_id(self,snapshot_id):
        snapshots=self.retrieve_snapshots()
        for snapshot in snapshots:
            if snapshot.attributes.id==snapshot_id:
                return snapshot
        return None 

    def retrieve_associated_volumes(self):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        self.update()
        volume_ids = self.attributes.volume_ids
        # build volume objects and add to list
        volume_objects = []
        for volume_id in volume_ids:
            newvolume = self.volume_manager.retrieve_volume_by_id(volume_id)
            volume_objects.append(newvolume)
        return volume_objects

    def count_associated_volumes(self):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        volumes = self.retrieve_associated_volumes()
        return len(volumes)

    def retrieve_associated_volume_snapshots(self):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        self.update() 
        volume_snapshot_ids = self.attributes.snapshot_ids
        # build volume objects and add to list
        volume_snapshot_objects = []
        for volume_snapshot_id in volume_snapshot_ids:
            newvolume_snapshot = self.snapshot_manager.retrieve_snapshot_id(
                volume_snapshot_id
            )
            volume_snapshot_objects.append(newvolume_snapshot)
        return volume_snapshot_objects

    def count_associated_volume_snapshots(self):
        if not self.deleted==False:
            raise ErrorDropletNotFound(f"{self.attributes.id} was deleted")
        volume_snapshots = self.retrieve_associated_volume_snapshots()
        return len(volume_snapshots)

    def attach_a_volume(self, target_volume:Volume):


        # Must check if the target_volume is in the same region
        if not self.attributes.region == target_volume.attributes.region:
            raise ErrorNotSameRegion(
                f"Volume {target_volume.attributes.id} not is same regions as Droplet {self.attributes.id}"
            )

        # Only 7 volumes allowed to be attached per droplet.
        self.update()
        if not self.count_associated_volumes() <=7 :
            raise ErrorDropletAttachedVolumeCountAlreadAtLimit(
                f"Droplet id:{self.attributes.id} already has the maximum of 7 attached volumes"
            )
        # Volumes can only be attached to one droplet
        ##remove volume from other droplets first
        target_volume.detach_from_droplets()

        response = self.volumeapi.attach_volume_to_droplet(
            target_volume.attributes.id, self.attributes.id,self.attributes.region['slug']
        )
        if response:
            content = json.loads(response.content.decode("utf-8"))
            action_data = dict(content["action"])
            newaction=Action(ActionAttributes(**action_data))
            self.action_manager.wait_for_action_completion(newaction)
            self.lastaction=newaction 
            self.update()
            target_volume.update()

    def detach_a_volume(self, target_volume:Volume):
        target_volume.detach_from_droplets()
        self.update()
        target_volume.update()




if __name__ == "__main__":
    dmanager = DropletManager()

    # a_droplet = dmanager.create_new_droplet(
    #    name="example.com",
    #    region="nyc3",
    #    size="s-1vcpu-1gb",
    #    image="ubuntu-16-04-x64",
    #    ssh_keys=[],
    #    backups=False,
    #    ipv6=True,
    #    user_data=None,
    #    private_networking=None,
    #    volumes=None,
    #    tags=["banabas"],
    # )

    # while not a_droplet.attributes.status == "active":
    #    time.sleep(5)
    #   print("waiting")

    # print("-----arguments-----")
    # print(a_droplet.arguments)
    # print("-------ATTRIBUTES-------------")
    # print(a_droplet.attributes)
    ## try:
    #    newdroplet = dmanager.retrieve_droplet_by_id(2496119371)
    # except DropletNotFound:
    #    print("Droplet wasnt found")
    # print(newdroplet.attributes)
    # print(newdroplet.arguments)
    # 249699371
    # droplets = dmanager.retrieve_all_droplets_by_tag("banabas")
    # for droplet in droplets:
    #    print(f"Deleteing droplet with id {droplet.attributes.id}")
    #    dmanager.delete_droplet(droplet)

    # print(dmanager.does_droplet_id_exist(249699371))

    # dmanager.delete_droplets_by_tag("banabas")
    droplets = dmanager.retrieve_droplets_by_name("example.com")
    for droplet in droplets:
        print(droplet.attributes.name)
        droplet.powercycle()
        while (droplet.lastaction.status == None) or (
            droplet.lastaction.status == "in-progress"
        ):
            time.sleep(5)
            print("waiting for action...")
        print(droplet.lastaction)
    print("finished")



from __future__ import annotations

from dataclasses import dataclass, field
from ..digitaloceanapi.volumes import Volumes
from ..digitaloceanapi.snapshots import Snapshots
from ..common.cloudapiexceptions import *
from .action import *
from .snapshot import *
from .account import *
import json
import threading
import time


@dataclass
class VolumeAttributes:
    id: str = None
    region: object = field(default_factory=list)
    droplet_ids: list = field(default_factory=list)
    name: str = None
    description: str = None
    size_gigabytes: int = None
    created_at: str = None
    filesystem_type: str = None
    filesystem_label: str = None
    tags: list = field(default_factory=list)


@dataclass
class VolumeArguments:
    size_gigabytes: int = None
    name: str = None
    region: str = None
    description: str = None
    snapshot_id: str = None
    filesystem_type: str = None
    filesystem_label: str = None
    tags: list = field(default_factory=list)


#@dataclass
#class VolumeLastActions:
#    id: int = None
#    status: str = None
#    type: str = None
#    started_at: str = None
#    completed_at: str = None
#    resource_id: int = None
#    resource_type: str = None
#    region: object = None
#    region_slug: str = None


class VolumeManager:
    def __init__(self):
        self.volumeapi = Volumes()
        self.account_manager = AccountManager()

    def check_limit(self):
        volume_limit = self.account_manager.volume_limit()
        if not len(self.retrieve_all_volumes()) < volume_limit:
            raise ErrorAccountVolumeLimitReached
            (
                f"You have reached your volume limit of {volume_limit}"
            )


    def create_new_volume(
        self,
        size_gigabytes: int,
        name: str,
        region: str,
        description: str = None,
        snapshot_id: str = None,
        filesystem_type: str = None,
        filesystem_label: str = None,
        tags: list= None,
    ):

        arguments = locals()
        del arguments["self"]

        if self.does_volume_name_region_exist(name, region):
            raise ErrorVolumeAlreadyExists(
                f"Volume name:{name}, region:{region} Already Exists"
            )

        self.check_limit()

        newvolume = Volume()
        newvolume.arguments = VolumeArguments(**arguments)
        response = self.volumeapi.create_new_volume(**arguments)
        if response:
            #
            volume_data = dict(json.loads(response.content.decode("utf-8"))["volume"])
            newvolume.attributes = VolumeAttributes(**volume_data)
        else:
            raise Exception(f"Could not create volume {name}")
        return newvolume

    def retrieve_all_volumes(self):

        volume_list = []
        page, per_page = 1, 10
        response = self.volumeapi.list_all_volumes(page=page, per_page=per_page)
        content = json.loads(response.content.decode("utf-8"))
        volume_list.extend(content["volumes"])
        try:
            while content["links"]["pages"]["next"]:
                page = page + 1
                response = self.volumeapi.list_all_volumes(page=page, per_page=per_page)
                content = json.loads(response.content.decode("utf-8"))
                volume_list.extend(content["volumes"])
        except KeyError:
            pass

        # Build and return that Volume object array.
        volume_objects = []
        for volume_item in volume_list:
            newvolume = Volume()
            newvolume.attributes = VolumeAttributes(**volume_item)
            newvolume.arguments = VolumeArguments()
            #You need to actually retreive the last action for the volume object before creating an Action
            #newvolume.lastaction = Action()
            volume_objects.append(newvolume)
        return volume_objects

    def retrieve_all_volumes_by_name(self, name):
        volume_list = []
        page, per_page = 1, 10
        response = self.volumeapi.list_all_volumes_by_name(
            name=name, page=page, per_page=per_page
        )
        content = json.loads(response.content.decode("utf-8"))
        volume_list.extend(content["volumes"])
        try:
            while content["links"]["pages"]["next"]:
                page = page + 1
                response = self.volumeapi.list_all_volumes_by_name(
                    name=name, page=page, per_page=per_page
                )
                content = json.loads(response.content.decode("utf-8"))
                volume_list.extend(content["volumes"])
        except KeyError:
            pass

        # Build and return that Volume object array.
        volume_objects = []
        for volume_item in volume_list:
            newvolume = Volume()
            newvolume.attributes = VolumeAttributes(**volume_item)
            newvolume.arguments = VolumeArguments()
            #You need to actually retreive the last action for the volume object before creating an Action
            #newvolume.lastaction = Action()
            volume_objects.append(newvolume)
        return volume_objects

    def retrieve_volume_by_id(self, id):
        if self.does_volume_id_exist(id):
            newvolume = Volume()
            response = self.volumeapi.retrieve_volume_by_id(id)
            content = json.loads(response.content.decode("utf-8"))
            volume_info = content["volume"]
            newvolume.attributes = VolumeAttributes(**volume_info)
            newvolume.arguments = VolumeArguments()
            #You need to actually retreive the last action for the volume object before creating an Action
            #newvolume.lastaction = Action()
            return newvolume
        else:
            raise ErrorVolumeNotFound(f"Volume {id} does not exist")

    def retrieve_volume_by_name_region(self, name, region):
        if self.does_volume_name_region_exist(name, region):
            newvolume = Volume()
            response = self.volumeapi.retrieve_volume_name_region(name, region)
            content = json.loads(response.content.decode("utf-8"))
            volume_info = content["volumes"][0]
            newvolume.attributes = VolumeAttributes(**volume_info)
            newvolume.arguments = VolumeArguments()
            #You need to actually retreive the last action for the volume object before creating an Action
            #newvolume.lastaction = Action()
            return newvolume
        else:
            raise ErrorVolumeNotFound(
                f"Volume name:{name}, region:{region} does not exist"
            )

    def retrieve_volumes_with_only_tags(self, tag: list):
        if isinstance(tag, str):
            tag = [tag]
        return_volumes = []
        volumes = self.retrieve_all_volumes()
        for volume in volumes:
            if set(list(tag)) == set(volume.attributes.tags):
                return_volumes.append(volume)
        if len(return_volumes) > 0:
            return return_volumes
        else:
            raise ErrorVolumeNotFound(
                f"No volumes containing all tags: {tag} were found."
            )

    def delete_volumes_with_only_tags(self, tag: list):
        if isinstance(tag, str):
            tag = [tag]
        try:
            volumes = self.retrieve_volumes_with_only_tags(tag)
            for volume in volumes:
                self.delete_volume(volume)
        except:
            pass

    def retrieve_volumes_with_all_tags(self, tag: list):
        if isinstance(tag, str):
            tag = [tag]
        return_volumes = []
        volumes = self.retrieve_all_volumes()
        for volume in volumes:
            if set(list(tag)).issubset(set(volume.attributes.tags)):
                return_volumes.append(volume)
        if len(return_volumes) > 0:
            return return_volumes
        else:
            raise ErrorVolumeNotFound(
                f"No volumes containing all tags: {tag} were found."
            )

    def delete_volumes_with_all_tags(self, tag: list):
        if isinstance(tag, str):
            tag = [tag]
        try:
            volumes = self.retrieve_volumes_with_all_tags(tag)
            for volume in volumes:
                self.delete_volume(volume)
        except:
            pass

    def retrieve_volumes_with_any_tags(self, tag: list):
        if isinstance(tag, str):
            tag = [tag]
        return_volumes = []
        volumes = self.retrieve_all_volumes()
        for volume in volumes:
            if not set(list(tag)).isdisjoint(set(volume.attributes.tags)):
                return_volumes.append(volume)
        if len(return_volumes) > 0:
            return return_volumes
        else:
            raise ErrorVolumeNotFound(
                f"No volumes containing any of the tags: {tag} were found."
            )

    def delete_volumes_with_any_tags(self, tag: list):
        if isinstance(tag, str):
            tag = [tag]
        try:
            volumes = self.retrieve_volumes_with_any_tags(tag)
            for volume in volumes:
                self.delete_volume(volume)
        except:
            pass

    def delete_volume(self, volume: Volume):
        """
        Given a Volume object, this method deletes the volume represnted by the object.

        Args:
            volume (volume): [description]
        """
        self.delete_volume_by_id(volume.attributes.id)

    def delete_volume_by_id(self, id):
        if self.does_volume_id_exist(id):
            response = self.volumeapi.delete_volume_id(id)

    def delete_volume_by_name_region(self, name=None, region=None):
        if (not name == None) and (not region == None):
            if self.does_volume_name_region_exist(name, region):
                self.volumeapi.delete_volume_name_region(name, region)

    def does_volume_id_exist(self, id):
        volumes = self.retrieve_all_volumes()
        for volume in volumes:
            if str(volume.attributes.id) == str(id):
                return True
        return False

    def does_volume_name_region_exist(self, name, region):
        volumes = self.retrieve_all_volumes()
        for volume in volumes:
            if (str(volume.attributes.name) == str(name)) and (
                str(volume.attributes.region["slug"]) == str(region)
            ):
                return True
        return False



class Volume:
    def __init__(self):
        self.arguments = VolumeArguments()
        self.attributes = VolumeAttributes()
        self.lastaction:Action = None
        self.volumeapi = Volumes()
        self.volume_manager = VolumeManager()
        self.action_manager = ActionManager()
        self.deleted=False
        
    def update(self):
        if not self.deleted==False:
            raise ErrorVolumeNotFound(f"{self.attributes.id} was deleted")
        """
        Updates the Volume attributes data class with the latest volume information at digital ocean.
        """
        response = self.volumeapi.retrieve_volume_by_id(self.attributes.id)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            volume_data = dict(content["volume"])
            self.attributes = VolumeAttributes(**volume_data)

    #Actions are self updating, no need for these methods.
    #def update_volume_action(self):
    #    """
    #    Updates the Droplet lastaction data class with the latest droplet action information at digital ocean.
    #    """
    #    response = self.volumeapi.retrieve_volume_action(
    #        self.attributes.id, self.lastaction.id
    #    )
    #    if response:
    #        content = json.loads(response.content.decode("utf-8"))
    #        action_data = dict(content["action"])
    #        self.lastaction = VolumeLastAction(**action_data)#
    #
    #def update_on_active_action(self):
    #    """
    #    A freshly created droplet will need time to completely boot up and be active.
    #    Information like IP addresses are not available untill the droplet is active.
    #    Here we start a background thread that waits for the droplet to become active and then update the droplet attributes.
    #    """#
    #
    #    def update_action():
    #        while (self.lastaction.status == None) or (
    #            self.lastaction.status == "in-progress"
    #        ):
    #            if not self.lastaction.status in ["completed", "errored"]:
    #                time.sleep(10)
    #                self.update_volume_action()
    #            else:
    #                break

    #    thread = threading.Thread(target=update_action, args=())
    #    thread.start()


    def create_snapshot(self, name, tags=[]):
        id = self.attributes.id
        if not self.volume_manager.does_volume_id_exist(id):
            raise ErrorVolumeNotFound(f"Volume id:{id} not found.")
        volumename = self.attributes.name
        arguments = {}
        arguments["name"] = name
        arguments["tags"] = tags
        if self.volume_manager.does_volume_id_exist(id):
            newsnapshot = Snapshot()
            newsnapshot.arguments = SnapshotArguments(**arguments)
            response = self.volumeapi.create_snapshot_from_volume(id, name, tags)
            if response:
                content = json.loads(response.content.decode("utf-8"))
                snapshot_info = content["snapshot"]
                newsnapshot.attributes = SnapshotAttributes(**snapshot_info)
                return newsnapshot

        else:
            raise ErrorVolumeNotFound(
                f"Cant create snapshot from non existent volume {id}:{volumename}"
            )

    def retrieve_snapshots(self):
        # Buildlist of snapshots from api, but take in to account pagination
        snapshot_list = []
        page, per_page = 1, 10
        response = self.volumeapi.list_snapshots_for_volume(
            id, page=page, per_page=per_page
        )
        content = json.loads(response.content.decode("utf-8"))
        print(content)
        snapshot_list.extend(content["snapshots"])
        try:
            while content["links"]["pages"]["next"]:
                response = self.volumeapi.list_snapshots_for_volume(
                    id, page=page, per_page=per_page
                )
                content = json.loads(response.content.decode("utf-8"))
                snapshot_list.extend(content["snapshots"])
        except KeyError:
            pass

        # Build and return that Snapshot object array.
        snapshot_objects = []
        for snapshot_item in snapshot_list:
            newsnapshot = Snapshot()
            newsnapshot.attributes = SnapshotAttributes(**snapshot_item)
            newsnapshot.arguments = SnapshotArguments()
            snapshot_objects.append(newsnapshot)
        return snapshot_objects



    

    def detach_from_droplets(self):
        #If volume is attached to any droplets, we detach it from them.
        self.update()
        droplet_ids=self.attributes.droplet_ids
        if len(droplet_ids)>0:
            for droplet_id in droplet_ids:
                response = self.volumeapi.detach_volume_from_droplet(
                self.attributes.id, droplet_id
            )
            if response:
                content = json.loads(response.content.decode("utf-8"))
                action_data = dict(content["action"])
                newaction=Action(ActionAttributes(**action_data))
                self.action_manager.wait_for_action_completion(newaction)
                self.lastaction=newaction

        

    def resize(self,size_gigabytes):
        #size from 1GB to max 16,384GB
        if(size_gigabytes>16384):
            raise ErrorVolumeResizeValueTooLarge("Maximum volume resize value of 16384 GB has been requested.")
        #can only resize upwards
        current_size=self.attributes.size_gigabytes
        target_size=size_gigabytes
        if(target_size<=current_size):
            raise ErrorVolumeResizeDirection("You tried to shrink a volume, this isn't allowed. You can only increase a volumes size")

        response=self.volumeapi.resize_volume(self.attributes.id,size_gigabytes,self.attributes.region['slug'])
        if response:
            content = json.loads(response.content.decode("utf-8"))
            action_data = dict(content["action"])
            newaction=Action(ActionAttributes(**action_data))
            self.action_manager.wait_for_action_completion(newaction)
            self.lastaction=newaction



from .digitaloceanapiconnection import DigitalOceanAPIConnection
import os
import time
import queue
import threading
import datetime
import random
import json


class FloatingIPs(DigitalOceanAPIConnection):
    def __init__(self):
        DigitalOceanAPIConnection.__init__(self)
        self.endpoint = "/v2/floating_ips"

    def list_all_floating_ips(self, page=0, per_page=0):
        arguments = locals()
        del arguments["self"]
        # params must be set from a dictionary not a json dump
        params = arguments
        return self.get_request(self.endpoint, headers=self.headers, params=params)

    def create_new_floating_ip(self, droplet_id):
        data_dict = {}
        data_dict["droplet_id"] = droplet_id
        data = json.dumps(data_dict)

        return self.post_request(self.endpoint, headers=self.headers, data=data)

    def reserve_ip_for_region(self, region):
        data_dict = {}
        data_dict["region"] = region
        data = json.dumps(data_dict)

        return self.post_request(self.endpoint, headers=self.headers, data=data)

    def delete_floating_ip(self, ip):
        self.delete_request(f"{self.endpoint}/{ip}")

    def assign_floating_ip_to_droplet(self, ip, droplet_id):
        data_dict = {}
        data_dict["type"] = "assign"
        data_dict["droplet_id"] = droplet_id

        data = json.dumps(data_dict)

        return self.post_request(
            f"{self.endpoint}/{ip}/actions", headers=self.headers, data=data
        )

    def unassign_floating_ip(self, ip):
        data_dict = {}
        data_dict["type"] = "unassign"

        data = json.dumps(data_dict)

        return self.post_request(
            f"{self.endpoint}/{ip}/actions", headers=self.headers, data=data
        )

    def list_all_actions(self, ip, page=0, per_page=0):
        # params must be set from a dictionary not a json dump
        params = {}
        params["page"] = page
        params["per_page"] = per_page

        return self.get_request(
            f"{self.endpoint}/{ip}/actions", headers=self.headers, params=params
        )



from .digitaloceanapiconnection import DigitalOceanAPIConnection
import os
import time
import queue
import threading
import datetime
import random
import json


class Volumes(DigitalOceanAPIConnection):
    """[summary]

    Args:
        DigitalOceanAPIConnection ([type]): [description]
    API Returns Volume Attributes:
        id(str):                The unique identifier for the block storage volume.
        region(object):         The region that the block storage volume is located in. When setting a region, the value should be the slug identifier for the region. When you query a block storage volume, the entire region object will be returned.
        droplet_ids([]):        An array containing the IDs of the Droplets the volume is attached to. Note that at this time, a volume can only be attached to a single Droplet.
        name(str):	            A human-readable name for the block storage volume. Must be lowercase and be composed only of numbers, letters and "-", up to a limit of 64 characters. The name must begin with a letter.
        description(str):	    An optional free-form text field to describe a block storage volume.
        size_gigabytes(int):	The size of the block storage volume in GiB (1024^3).
        created_at(str):	    A time value given in ISO8601 combined date and time format that represents when the block storage volume was created.
        filesystem_type(str):	The type of filesystem currently in-use on the volume.
        filesystem_label(str)   The label currently applied to the filesystem.
        tags([]):               An array of tags the volume has been tagged with

    """

    def __init__(self):
        DigitalOceanAPIConnection.__init__(self)
        self.endpoint = "/v2/volumes"

    def list_all_volumes(self, page=0, per_page=0):
        arguments = locals()
        del arguments["self"]
        # params must be set from a dictionary not a json dump
        params = arguments

        return self.get_request(self.endpoint, headers=self.headers, params=params)

    def list_all_volumes_by_name(self, name, page=0, per_page=0):
        arguments = locals()
        del arguments["self"]
        # params must be set from a dictionary not a json dump
        params = arguments

        return self.get_request(self.endpoint, headers=self.headers, params=params)

    def create_new_volume(
        self,
        size_gigabytes: int,
        name: str,
        region: str,
        description: str = None,
        snapshot_id: str = None,
        filesystem_type: str = None,
        filesystem_label: str = None,
        tags: list = None,
    ):

        """[summary]

        Returns:
            [type]: [description]
        API Expects:
                size_gigabytes(int):	REQUIRED The size of the block storage volume in GiB (1024^3).
                name(str):	            REQUIRED A human-readable name for the block storage volume. Must be lowercase and be composed only of numbers, letters and "-", up to a limit of 64 characters.
                region(str):	        REQUIRED The region where the block storage volume will be created. When setting a region, the value should be the slug identifier for the region. When you query a block storage volume, the entire region object will be returned.
                description(str):	    string	An optional free-form text field to describe a block storage volume.
                snapshot_id(str):       The unique identifier for the volume snapshot from which to create the volume. Should not be specified with a region_id.
                filesystem_type(str):	The name of the filesystem type to be used on the volume. When provided, the volume will automatically be formatted to the specified filesystem type. Currently, the available options are "ext4" and "xfs". Pre-formatted volumes are automatically mounted when attached to Ubuntu, Debian, Fedora, Fedora Atomic, and CentOS Droplets created on or after April 26, 2018. Attaching pre-formatted volumes to other Droplets is not recommended.
                filesystem_label(str):	The label to be applied to the filesystem. Labels for ext4 type filesystems may contain 16 characters while lables for xfs type filesystems are limited to 12 characters. May only be used in conjunction with filesystem_type.
                tags([]):   	        A flat array of tag names as strings to apply to the Volume after it is created. Tag names can either be existing or new tags.


        API data example:


        """
        arguments = locals()
        del arguments["self"]
        data = json.dumps(arguments)
        return self.post_request(f"{self.endpoint}", headers=self.headers, data=data)

    def retrieve_volume_by_id(self, id):
        return self.get_request(f"{self.endpoint}/{id}", headers=self.headers)

    def retrieve_volume_name_region(self, name, region):
        arguments = locals()
        del arguments["self"]
        # params must be set from a dictionary not a json dump
        params = arguments
        return self.get_request(self.endpoint, headers=self.headers, params=params)

    def delete_volume_id(self, id):
        return self.delete_request(f"{self.endpoint}/{id}", headers=self.headers)

    def delete_volume_name_region(self, name, region):
        arguments = locals()
        del arguments["self"]
        # params must be set from a dictionary not a json dump
        params = arguments
        return self.delete_request(self.endpoint, headers=self.headers, params=params)

    def create_snapshot_from_volume(self, id, name, tags=[]):
        data_dict = {}
        data_dict["name"] = name
        data_dict["tags"] = tags
        data = json.dumps(data_dict)
        return self.post_request(
            f"{self.endpoint}/{id}/snapshots", headers=self.headers, data=data
        )

    def list_snapshots_for_volume(self, id, page=0, per_page=0):
        arguments = locals()
        del arguments["self"]
        del arguments["id"]
        # params must be set from a dictionary not a json dump
        params = arguments
        return self.get_request(
            f"{self.endpoint}/{id}/snapshots", headers=self.headers, params=params
        )

    def attach_volume_to_droplet(self, volume_id, droplet_id, region, tags=[]):
        data_dict = {}
        data_dict["type"] = "attach"
        data_dict["droplet_id"] = droplet_id
        data_dict["region"] = str(region)
        data_dict["tags"] = tags
        data = json.dumps(data_dict)
        return self.post_request(
            f"{self.endpoint}/{volume_id}/actions", headers=self.headers, data=data
        )

    def detach_volume_from_droplet(self, volume_id, droplet_id, region=None):
        data_dict = {}
        data_dict["type"] = "detach"
        data_dict["droplet_id"] = droplet_id
        if not region == None:
            data_dict["region"] = region
        data = json.dumps(data_dict)
        return self.post_request(
            f"{self.endpoint}/{volume_id}/actions", headers=self.headers, data=data
        )

    def retrieve_volume_action(self, volume_id, action_id):

        return self.get_request(
            f"{self.endpoint}/{volume_id}/actions/{action_id}", headers=self.headers
        )

    def resize_volume(self, volume_id, size, region=None):
        # size from 1GB to max 16,384GB
        data_dict = []
        data_dict["type"] = "resize"
        data_dict["size_gigabytes"] = size
        if not region == None:
            data_dict["region"] = region
        data = json.dumps(data_dict)

        return self.post_request(
            f"{self.endpoint}/{volume_id}/actions", headers=self.headers, data=data
        )



from __future__ import annotations


from ..digitaloceanapi.accounts import Accounts
from ..common.cloudapiexceptions import *

from dataclasses import dataclass, field

from requests.models import Response
import json
import threading
import time
import re


@dataclass
class AccountAttributes:
    droplet_limit: int = None
    floating_ip_limit: int = None
    volume_limit: int = None
    email: str = None
    uuid: str = None
    email_verified: bool = None
    status: str = None
    status_message: str = None


class AccountManager:
    def __init__(self):
        self.accountapi = Accounts()
        self.account_details = None

    def retrieve_account_details(self):
        response = self.accountapi.list_account_information()
        if response:
            content = json.loads(response.content.decode("utf-8"))
            account_data = content["account"]
            newaccount = Account()
            newaccount.attributes = AccountAttributes(**account_data)
            print(newaccount.attributes)
            return newaccount

    def droplet_limit(self):
        self.account_details = self.retrieve_account_details()
        return self.account_details.attributes.droplet_limit

    def floating_ip_limit(self):
        self.account_details = self.retrieve_account_details()
        return self.account_details.attributes.floating_ip_limit

    def volume_limit(self):
        self.account_details = self.retrieve_account_details()
        return self.account_details.attributes.volume_limit

    def email(self):
        self.account_details = self.retrieve_account_details()
        return self.account_details.attributes.email

    def uuid(self):
        self.account_details = self.retrieve_account_details()
        return self.account_details.attributes.uuid

    def email_verified(self):
        self.account_details = self.retrieve_account_details()
        return self.account_details.attributes.email_verified

    def status(self):
        self.account_details = self.retrieve_account_details()
        return self.account_details.attributes.status

    def status_message(self):
        self.account_details = self.retrieve_account_details()
        return self.account_details.attributes.status_message


class Account:
    def __init__(self, status=None):
        self.attributes = AccountAttributes()



from __future__ import annotations

from dataclasses import dataclass, field
from ..digitaloceanapi.actions import Actions
from ..common.cloudapiexceptions import *
import json
import threading
import time
import sys


@dataclass
class ActionAttributes:
    id: int = None
    status: str = None
    type: str = None
    started_at: str = None
    completed_at: str = None
    resource_id: int = None
    resource_type: str = None
    region: object = None
    region_slug: str = None


class ActionManager:
    def __init__(self):
        self.actionapi = Actions()

    def retrieve_all_actions(self):
        """
        Returns an array of Droplet objects, one for each droplet in digitalocean account.

        Returns:
            [type]: [description]
        """

        # Build list of droplets from api, but take in to account possible pagination.
        action_list = []
        page, per_page = 1, 10
        response = self.actionapi.list_all_actions(page=page, per_page=per_page)
        content = json.loads(response.content.decode("utf-8"))
        action_list.extend(content["actions"])
        try:
            while content["links"]["pages"]["next"]:
                page = page + 1
                # if page % 10 == 0:
                #    print(page)
                response = self.actionapi.list_all_actions(page=page, per_page=per_page)
                content = json.loads(response.content.decode("utf-8"))
                action_list.extend(content["actions"])
                sys.stdout.flush()
        except KeyError:
            pass

        # Build and return that Droplet object array.
        action_objects = []
        for action_item in action_list:
            newaction = Action()
            newaction.attributes = ActionAttributes(**action_item)
            action_objects.append(newaction)
        return action_objects

    def retrieve_paginated_actions(self, start_page, end_page, per_page=10):
        """
        Returns an array of Droplet objects, one for each droplet in digitalocean account.

        Returns:
            [type]: [description]
        """

        # Build list of droplets from api, but take in to account possible pagination.
        action_list = []
        page = start_page
        response = self.actionapi.list_all_actions(page=page, per_page=per_page)
        content = json.loads(response.content.decode("utf-8"))
        action_list.extend(content["actions"])
        try:
            while content["links"]["pages"]["next"]:
                page = page + 1
                response = self.actionapi.list_all_actions(page=page, per_page=per_page)
                content = json.loads(response.content.decode("utf-8"))
                action_list.extend(content["actions"])
                sys.stdout.flush()
                if page >= end_page:
                    break
        except KeyError:
            pass

        # Build and return that Droplet object array.
        action_objects = []
        for action_item in action_list:
            newaction = Action()
            newaction.attributes = ActionAttributes(**action_item)
            action_objects.append(newaction)
        return action_objects

    def does_action_exist_id(self, action_id):
        start_page, end_page = 1, 10
        actions = self.retrieve_paginated_actions(
            start_page=start_page, end_page=end_page
        )
        for action in actions:
            if action.attributes.id == action_id:
                return True
        full_result_count = len(actions)
        while True:
            start_page, end_page = start_page + 10, end_page + 10
            actions = self.retrieve_paginated_actions(
                start_page=start_page, end_page=end_page
            )
            for action in actions:
                if action.attributes.id == action_id:
                    print(action.attributes)
                    return True
            if len(actions) < full_result_count:
                break
        return False

    def retrive_action(self, action_id, check_if_exists=True):
        if check_if_exists:
            if not self.does_action_exist_id(action_id):
                raise ErrorActionDoesNotExists(
                    f"action with id:{action_id} does not exist"
                )
        response = self.actionapi.retrieve_existing_action(action_id)
        content = json.loads(response.content.decode("utf-8"))
        action = Action()
        action.attributes = ActionAttributes(**content["action"])
        return action

    def wait_for_action_completion(self, action: Action):
        while not action.attributes.status in ["completed", "errored"]:
            time.sleep(5)
        if action.attributes.status == "errored":
            raise ErrorActionFailed(
                f"Action {self.attributes.id},{self.attributes.type} failed"
            )


class Action:
    def __init__(self, action_attributes: ActionAttributes):
        self.attributes = action_attributes
        self.actionapi = Actions()
        self.update_on_active_action()

    def update_action_action(self):
        """
        Updates the Droplet lastaction data class with the latest droplet action information at digital ocean.
        """
        response = self.actionapi.retrieve_existing_action(self.attributes.id)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            action_data = dict(content["action"])
            self.attributes = ActionAttributes(**action_data)

    def update_on_active_action(self):
        def update_action():
            # On creating a blank action it's id is None before a user defines an action.
            if self.attributes.id == None:
                pass
            while (self.attributes.status == None) or (
                self.attributes.status == "in-progress"
            ):
                if not self.attributes.status in ["completed", "errored"]:
                    time.sleep(10)
                    self.update_action_action()
                else:
                    break
                if self.attributes.status == "errored":
                    raise ErrorActionFailed(
                        f"Action {self.attributes.id},{self.attributes.type} failed"
                    )

        thread = threading.Thread(target=update_action, args=())
        thread.start()



from __future__ import annotations

from dataclasses import dataclass, field
from ..digitaloceanapi.sizes import Sizes
from ..common.cloudapiexceptions import *
import json
import threading
import time
import sys


@dataclass
class SizeAttributes:
    slug: str = None
    available: bool = None
    transfer: float = None
    price_monthly: float = None
    price_hourly: float = None
    memory: int = None
    vcpus: int = None
    disk: int = None
    regions: list = field(default_factory=list)
    description: str = None


class SizeManager:
    def __init__(self):
        self.sizeapi = Sizes()

    def retrieve_sizes(self):
        response = self.sizeapi.list_all_sizes()
        size_list = []
        return_sizes = []
        if response:
            content = json.loads(response.content.decode("utf-8"))
            size_list = content["sizes"]
        for size_data in size_list:
            newsize = Size()
            newsize.attributes = SizeAttributes(**size_data)
            return_sizes.append(newsize)
        return return_sizes

    def retrieve_size(self, slug):
        sizes = self.retrieve_sizes()
        for size in sizes:
            if size.attributes.slug == slug:
                return size
        raise ErrorDropletSlugSizeNotFound(f'"{slug}" not found')


class Size:
    def __init__(self):
        self.attributes = SizeAttributes()



from __future__ import annotations

from dataclasses import dataclass, field
from ..digitaloceanapi.droplets import Droplets
from ..digitaloceanapi.floatingips import FloatingIPs
from .action import *
from .droplet import *
from .region import *
from .account import *
from ..common.cloudapiexceptions import *
import json
import threading
import time
import re


@dataclass
class FloatingIPAttributes:
    ip: str = None
    region: object = None
    droplet: object = None
    locked: bool = None


class FloatingIPManager:
    def __init__(self):
        self.floatingipapi = FloatingIPs()
        self.action_manager = ActionManager()
        self.region_manager = RegionManager()
        self.account_manager = AccountManager()

    def check_limit(self):
        floating_ip_limit = self.account_manager.floating_ip_limit()
        if not len(self.retrieve_all_floating_ips()) < floating_ip_limit:
            raise ErrorAccountFloatingIPLimitReached(
                f"You have reached your floating ip limit of {floating_ip_limit}"
            )

    def retrieve_all_floating_ips(self):
        floating_ip_list = []
        page, per_page = 1, 10
        response = self.floatingipapi.list_all_floating_ips(
            page=page, per_page=per_page
        )
        content = json.loads(response.content.decode("utf-8"))
        floating_ip_list.extend(content["floating_ips"])
        try:
            while content["links"]["pages"]["next"]:
                page = page + 1
                response = self.floatingipapi.list_all_floating_ips(
                    page=page, per_page=per_page
                )
                content = json.loads(response.content.decode("utf-8"))
                floating_ip_list.extend(content["floating_ips"])
        except KeyError:
            pass

        floating_ip_objects = []
        for floating_ip in floating_ip_list:
            newfloatingip = FloatingIP()
            newfloatingip.attributes = FloatingIPAttributes(**floating_ip)
            floating_ip_objects.append(newfloatingip)
        return floating_ip_objects

    def create_new_floating_ip(self, droplet: Droplet):
        self.check_limit()
        floatingip = self.check_droplet_for_floating_ip(droplet)
        if not floatingip == None:
            raise ErrorDropletAlreadyHasFloatingIP(
                f"Droplet id:{droplet.attributes.id} already has a private ip: {floatingip.attributes.ip}"
            )

        response = self.floatingipapi.create_new_floating_ip(droplet.attributes.id)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            floating_ip_data = content["floating_ip"]
            newfloatingip = FloatingIP()
            newfloatingip.attributes = FloatingIPAttributes(**floating_ip_data)
            return newfloatingip

    def check_droplet_for_floating_ip(self, droplet: Droplet):
        floating_ips = self.retrieve_all_floating_ips()
        for floating_ip in floating_ips:
            if not floating_ip.attributes.droplet == None:
                if floating_ip.attributes.droplet["id"] == droplet.attributes.id:
                    return floating_ip
        return None

    def reserve_ip_for_region(self, region_slug):
        self.check_limit()
        if self.region_manager.does_region_exist(region_slug):
            response = self.floatingipapi.reserve_ip_for_region(region_slug)
            if response:
                content = json.loads(response.content.decode("utf-8"))
                floating_ip_data = content["floating_ip"]
                newfloatingip = FloatingIP()
                newfloatingip.attributes = FloatingIPAttributes(**floating_ip_data)
                return newfloatingip
        else:
            raise ErrorRegionDoesNotExist(f'"{region_slug}" not a valid region')

    def retrieve_floating_ip(self, ip):
        floatingips = self.retrieve_all_floating_ips()
        for floatingip in floatingips:
            if floatingip.attributes.ip == ip:
                return floatingip
        raise ErrorFloatingIPDoesNotExists(
            f"Could not find ip:{ip} associated with your account"
        )


class FloatingIP:
    def __init__(self):
        self.attributes = FloatingIPAttributes()
        self.lastaction: Action = None
        self.floatingipapi = FloatingIPs()
        self.floatingip_manager = FloatingIPManager()
        self.action_manager = ActionManager()

    def delete(self):
        self.floatingipapi.delete_floating_ip(self.attributes.ip)

    def unassign(self):
        newaction = Action()
        response = self.floatingipapi.unassign_floating_ip(self.attributes.ip)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            newaction.attributes = ActionAttributes(content["action"])
            self.action_manager.wait_for_action_completion(newaction)

    def attach_to_droplet(self, droplet: Droplet):
        if not self.floatingip_manager.check_droplet_for_floating_ip(droplet) == None:
            self.unassign()
        newaction = Action()
        response = self.floatingipapi.assign_floating_ip_to_droplet(
            self.attributes.ip, droplet.attributes.id
        )
        if response:
            content = json.loads(response.content.decode("utf-8"))
            newaction.attributes = ActionAttributes(content["action"])
            self.action_manager.wait_for_action_completion(newaction)

    def retrieve_all_actions(self):
        action_list = []
        page, per_page = 1, 10
        response = self.floatingipapi.list_all_actions(page=page, per_page=per_page)
        content = json.loads(response.content.decode("utf-8"))
        action_list.extend(content["actions"])
        try:
            while content["links"]["pages"]["next"]:
                page = page + 1
                response = self.floatingipapi.list_all_actions(
                    page=page, per_page=per_page
                )
                content = json.loads(response.content.decode("utf-8"))
                action_list.extend(content["actions"])
        except KeyError:
            pass

        # Build and return that Droplet object array.
        action_objects = []
        for action_item in action_list:
            newaction = Action()
            newaction.attributes = ActionAttributes(**action_item)
            action_objects.append(newaction)
        return action_objects

    def retrieve_existing_action(self, action_id):
        floatingip_actions = self.retrieve_existing_action()
        for floatingip_action in floatingip_actions:
            if floatingip_action.attributes.id == action_id:
                return floatingip_action
        raise ErrorActionDoesNotExists(
            f"floating ip:{self.attributes.ip} does not have associated actions"
        )



from __future__ import annotations

from dataclasses import dataclass, field
from ..digitaloceanapi.sshkeys import SSHkeys
from ..common.cloudapiexceptions import *
import json
import threading
import time
import re


@dataclass
class SSHkeyAttributes:
    id: str = None
    fingerprint: str = None
    public_key: str = None
    name: str = None


class SSHkeyManager:
    def __init__(self):
        self.sshkeyapi = SSHkeys()

    def retrieve_all_sshkeys(self):
        sshkey_objects = []
        response = self.sshkeyapi.list_all_keys()
        if response:
            content = json.loads(response.content.decode("utf-8"))
            sshkey_datas = content["ssh_keys"]
            for sshkey_data in sshkey_datas:
                newsshkey = SSHkey()
                newsshkey.attributes = SSHkeyAttributes(**sshkey_data)
                sshkey_objects.append(newsshkey)
        return sshkey_objects

    def create_new_key(self, name, public_key):
        newsshkey = SSHkey()
        response = self.sshkeyapi.create_new_key(name, public_key)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            sshkey_data = content["ssh_key"]
            newsshkey = SSHkey()
            newsshkey.attributes = SSHkeyAttributes(**sshkey_data)
            return newsshkey

    def retrieve_sshkey_with_id(self, id):
        sshkey_objects = self.retrieve_all_sshkeys()
        for sshkey in sshkey_objects:
            if sshkey.attributes.id == id:
                return sshkey
        raise ErrorSSHkeyDoesNotExists(f"SSHkey with id {id} could not be found.")

    def does_sshkey_exist_id(self, id):
        sshkey_objects = self.retrieve_all_sshkeys()
        for sshkey in sshkey_objects:
            if sshkey.attributes.id == id:
                return True
        return False


class SSHkey:
    def __init__(self):
        self.attributes = SSHkeyAttributes()
        self.sshkeyapi = SSHkeys()
        self.sshkey_manager = SSHkeyManager()

    def update_name(self, name):
        if self.sshkey_manager.does_sshkey_exist_id(self.attributes):
            response = self.sshkeyapi.update_name(self.attributes.id, name)
            if response:
                content = json.loads(response.content.decode("utf-8"))
                sshkey_data = content["ssh_key"]
                self.attributes = SSHkeyAttributes(**sshkey_data)
        raise ErrorSSHkeyDoesNotExists(
            f"SSHkey with id {self.attributes.id} could not be found."
        )

    def delete(self):
        if self.sshkey_manager.does_sshkey_exist_id(self.attributes):
            self.sshkeyapi.delete_sshkey(self.attributes.id)



from .digitaloceanapiconnection import DigitalOceanAPIConnection
import os
import time
import queue
import threading
import datetime
import random
import json


class Regions(DigitalOceanAPIConnection):
    def __init__(self):
        DigitalOceanAPIConnection.__init__(self)
        self.endpoint = "/v2/regions"

    def list_all_regions(self):
        return self.get_request(self.endpoint, headers=self.headers)



from __future__ import annotations

from dataclasses import dataclass, field
from ..digitaloceanapi.regions import Regions

from ..common.cloudapiexceptions import *
import json
import threading
import time
import re


@dataclass
class RegionAttributes:
    slug: str = None
    name: str = None
    sizes: list = field(default_factory=list)
    available: bool = None
    features: list = field(default_factory=list)


class RegionManager:
    def __init__(self):
        self.regionapi = Regions()

    def retrieve_all_regions(self):
        region_objects = []
        response = self.regionapi.list_all_regions()
        if response:
            content = json.loads(response.content.decode("utf-8"))
            region_datas = content["regions"]
            for region_data in region_datas:
                newregion = Region()
                newregion.attributes = RegionAttributes(**region_data)
                region_objects.append(newregion)
            return region_objects

    def does_region_exist(self, region_slug):
        digital_ocean_regions = self.retrieve_all_regions()
        for digital_ocean_region in digital_ocean_regions:
            if digital_ocean_region.attributes.slug == region_slug:
                return True
        return False


class Region:
    def __init__(self):
        self.attributes = RegionAttributes()



from __future__ import annotations

from dataclasses import dataclass, field
from ..common.cloudapiexceptions import *
from ..digitaloceanapi.snapshots import Snapshots
import json
import threading
import time
import datetime


class SnapshotManager:
    def __init__(self):
        self.snapshotapi = Snapshots()

    def retrieve_all_snapshots(self):
        snapshot_list = []
        page, per_page = 1, 10
        response = self.snapshotapi.list_all_snapshots(page=page, per_page=per_page)
        content = json.loads(response.content.decode("utf-8"))
        snapshot_list.extend(content["snapshots"])
        try:
            while content["links"]["pages"]["next"]:
                page = page + 1
                response = self.snapshotapi.list_all_snapshots(
                    page=page, per_page=per_page
                )
                content = json.loads(response.content.decode("utf-8"))
                snapshot_list.extend(content["snapshots"])
        except KeyError:
            pass

        # Build and return that Snapshot object array.
        snapshot_objects = []
        for snapshot_item in snapshot_list:
            newsnapshot = Snapshot()
            newsnapshot.attributes = SnapshotAttributes(**snapshot_item)
            newsnapshot.arguments = SnapshotArguments()
            snapshot_objects.append(newsnapshot)
        return snapshot_objects

    def retrieve_all_droplet_snapshots(self):
        snapshot_list = []
        page, per_page = 1, 10
        response = self.snapshotapi.list_all_droplet_snapshots(
            page=page, per_page=per_page
        )
        content = json.loads(response.content.decode("utf-8"))
        snapshot_list.extend(content["snapshots"])
        try:
            while content["links"]["pages"]["next"]:
                page = page + 1
                response = self.snapshotapi.list_all_droplet_snapshots(
                    page=page, per_page=per_page
                )
                content = json.loads(response.content.decode("utf-8"))
                snapshot_list.extend(content["snapshots"])
        except KeyError:
            pass

        # Build and return that Snapshot object array.
        snapshot_objects = []
        for snapshot_item in snapshot_list:
            newsnapshot = Snapshot()
            newsnapshot.attributes = SnapshotAttributes(**snapshot_item)
            newsnapshot.arguments = SnapshotArguments()
            snapshot_objects.append(newsnapshot)
        return snapshot_objects

    def retrieve_all_volume_snapshots(self):
        snapshot_list = []
        page, per_page = 1, 10
        response = self.snapshotapi.list_all_volume_snapshots(
            page=page, per_page=per_page
        )
        content = json.loads(response.content.decode("utf-8"))
        snapshot_list.extend(content["snapshots"])
        try:
            while content["links"]["pages"]["next"]:
                page = page + 1
                response = self.snapshotapi.list_all_volume_snapshots(
                    page=page, per_page=per_page
                )
                content = json.loads(response.content.decode("utf-8"))
                snapshot_list.extend(content["snapshots"])
        except KeyError:
            pass

        # Build and return that Snapshot object array.
        snapshot_objects = []
        for snapshot_item in snapshot_list:
            newsnapshot = Snapshot()
            newsnapshot.attributes = SnapshotAttributes(**snapshot_item)
            newsnapshot.arguments = SnapshotArguments()
            snapshot_objects.append(newsnapshot)
        return snapshot_objects

    def retrieve_snapshot_id(self, id):
        if not self.does_snapshot_id_exist(id):
            raise ErrorSnapshotNotFound(f"Snapshot with id:{id} not found")
        newsnapshot = Snapshot()
        response = self.snapshotapi.retrieve_snapshot_by_id(id)
        if response:
            content = json.loads(response.content.decode("utf-8"))
            snapshot_data = dict(content["snapshots"])
            newsnapshot.attributes = SnapshotAttributes(**snapshot_data)
        return newsnapshot

    def delete_snapshot(self, snapshot: Snapshot):
        id = snapshot.attributes.id
        self.delete_snapshot_id(id)

    def delete_snapshot_id(self, id):
        if self.does_snapshot_id_exist(id):
            self.snapshotapi.delete_snapshot_id(id)

    def does_snapshot_id_exist(self, id):
        snapshots = self.retrieve_all_snapshots()
        for snapshot in snapshots:
            if str(snapshot.attributes.id) == str(id):
                return True
        return False


@dataclass
class SnapshotArguments:
    name: str = None
    tags: list = field(default_factory=list)


@dataclass
class SnapshotAttributes:
    id: str = None
    name: str = None
    created_at: str = None
    regions: list = field(default_factory=list)
    resource_id: str = None
    resource_type: str = None
    min_disk_size: int = None
    size_gigabytes: float = None
    tags: list = field(default_factory=list)


class Snapshot:
    def __init__(self):
        self.arguments = SnapshotArguments()
        self.attributes = SnapshotAttributes()



# from distutils.core import setup
##How to package
##From your root package dir...
##python3 setup.py sdist
##twine check dist/*
##twine upload dist/*


from setuptools import setup

# read the contents of your README file
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="digitaloceanobjects",
    packages=[
        "digitaloceanobjects",
        "digitaloceanobjects.common",
        "digitaloceanobjects.digitaloceanapi",
        "digitaloceanobjects.digitaloceanobject",
    ],
    version="0.0.15",
    license="MIT",
    description="digitaloceanobjects, represents all digital ocean services as objects, hiding all those horrible api calls.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="zoran ilievski",
    author_email="pythonic@clientuser.net",
    url="https://github.com/zorani/digitalocean-objects",
    download_url="https://github.com/zorani/digitalocean-objects/archive/refs/tags/v0.0.15.tar.gz",
    keywords=["digitalocean", "api", "objects"],
    install_requires=["cloudapi"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



from .digitaloceanobject.droplet import Droplet as Droplet
from .digitaloceanobject.droplet import DropletManager as DropletManager
from .digitaloceanobject.volume import Volume as Volume
from .digitaloceanobject.volume import VolumeManager as VolumeManager
from .digitaloceanobject.snapshot import Snapshot as Snapshot
from .digitaloceanobject.snapshot import SnapshotManager as SnapshotManager
from .digitaloceanobject.action import Action as Action
from .digitaloceanobject.action import ActionManager as ActionManager
from .digitaloceanobject.size import Size as Size
from .digitaloceanobject.size import SizeManager as SizeManager
from .digitaloceanobject.floatingip import FloatingIP as FloatingIP
from .digitaloceanobject.floatingip import FloatingIPManager as FloatingIPManager
from .digitaloceanobject.account import Account as Account
from .digitaloceanobject.account import AccountManager as AccountManager
from .digitaloceanobject.sshkey import SSHkey as SSHkey
from .digitaloceanobject.sshkey import SSHkeyManager as SSHkeyManager
from .digitaloceanobject.account import Account as Account
from .digitaloceanobject.account import AccountManager as AccountManager
from .digitaloceanobject.region import Regions as Region
from .digitaloceanobject.region import RegionManager as RegionManager
from .common.cloudapiexceptions import ErrorDropletNotFound as ErrorDropletNotFound
from .common.cloudapiexceptions import (
    ErrorVolumeAlreadyExists as ErrorVolumeAlreadyExists,
)
from .common.cloudapiexceptions import ErrorVolumeNotFound as ErrorVolumeNotFound
from .common.cloudapiexceptions import (
    ErrorVolumeResizeValueTooLarge as ErrorVolumeResizeValueTooLarge,
)
from .common.cloudapiexceptions import (
    ErrorVolumeResizeDirection as ErrorVolumeResizeDirection,
)
from .common.cloudapiexceptions import ErrorSnapshotNotFound as ErrorSnapshotNotFound
from .common.cloudapiexceptions import (
    ErrorDropletNameContainsSpaces as ErrorDropletNameContainsSpaces,
)
from .common.cloudapiexceptions import (
    ErrorActionDoesNotExists as ErrorActionDoesNotExists,
)
from .common.cloudapiexceptions import ErrorActionFailed as ErrorActionFailed
from .common.cloudapiexceptions import (
    ErrorDropletSlugSizeNotFound as ErrorDropletSlugSizeNotFound,
)
from .common.cloudapiexceptions import (
    ErrorDropletNameContainsInvalidChars as ErrorDropletNameContainsInvalidChars,
)
from .common.cloudapiexceptions import (
    ErrorDropletResizeDiskError as ErrorDropletResizeDiskError,
)
from .common.cloudapiexceptions import (
    ErrorDropletAttachedVolumeCountAlreadAtLimit as ErrorDropletAttachedVolumeCountAlreadAtLimit,
)
from .common.cloudapiexceptions import ErrorNotSameRegion as ErrorNotSameRegion
from .common.cloudapiexceptions import (
    ErrorDropletAlreadyHasFloatingIP as ErrorDropletAlreadyHasFloatingIP,
)
from .common.cloudapiexceptions import (
    ErrorRegionDoesNotExist as ErrorRegionDoesNotExist,
)
from .common.cloudapiexceptions import (
    ErrorAccountFloatingIPLimitReached as ErrorAccountFloatingIPLimitReached,
)
from .common.cloudapiexceptions import (
    ErrorAccountDropletLimitReached as ErrorAccountDropletLimitReached,
)
from .common.cloudapiexceptions import (
    ErrorAccountVolumeLimitReached as ErrorAccountVolumeLimitReached,
)
from .common.cloudapiexceptions import (
    ErrorFloatingIPDoesNotExists as ErrorFloatingIPDoesNotExists,
)
from .common.cloudapiexceptions import (
    ErrorSSHkeyDoesNotExists as ErrorSSHkeyDoesNotExists,
)



class ErrorDropletNotFound(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorVolumeAlreadyExists(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorVolumeNotFound(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorVolumeResizeValueTooLarge(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorVolumeResizeDirection(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorSnapshotNotFound(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorDropletNameContainsSpaces(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorActionDoesNotExists(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorActionFailed(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorDropletSlugSizeNotFound(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorDropletNameContainsInvalidChars(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorDropletResizeDiskError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorDropletAttachedVolumeCountAlreadAtLimit(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorNotSameRegion(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorDropletAlreadyHasFloatingIP(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorRegionDoesNotExist(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorAccountFloatingIPLimitReached(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorAccountDropletLimitReached(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorAccountVolumeLimitReached(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorFloatingIPDoesNotExists(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ErrorSSHkeyDoesNotExists(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)



from .digitaloceanapiconnection import DigitalOceanAPIConnection
import os
import time
import queue
import threading
import datetime
import random
import json


class SSHkeys(DigitalOceanAPIConnection):
    def __init__(self):
        DigitalOceanAPIConnection.__init__(self)
        self.endpoint = "/v2/account/keys"

    def list_all_keys(self):
        return self.get_request(self.endpoint, headers=self.headers)

    def create_new_key(self, name, public_key):
        data_dict = {}
        data_dict["name"] = name
        data_dict["public_key"] = public_key
        data = json.dumps(data_dict)

        return self.post_request(self.endpoint, headers=self.headers, data=data)

    def update_name(self, id, name):
        data_dict = {}
        data_dict["name"] = name
        data = json.dumps(data_dict)

        return self.put_request(
            f"{self.endpoint}/{id}", headers=self.headers, data=data
        )

    def delete_sshkey(self, identifier):
        # identifier could be the sshkey id or the sshkey fingerprint
        self.delete_request(f"{self.endpoint}/{identifier}")
